#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the search related array."""

from typing import Any, Callable, Dict, Union

import graviti.portex as pt
from graviti.dataframe.sql.container import (
    SearchContainer,
    SearchContainerRegister,
    SearchScalarContainer,
)
from graviti.dataframe.sql.scalar import (
    BooleanScalar,
    ComparisonArithmeticOperatorsMixin,
    EnumScalar,
    NumberScalar,
    RowSeries,
    StringScalar,
)


class LogicalOperatorsMixin:
    """A mixin for dynamically implementing logical operators."""

    expr: Union[str, Dict[str, Any]]
    schema: pt.PortexType
    logical_operators: Dict[str, str] = {
        "__eq__": "eq",
        "__ne__": "ne",
        "__and__": "and",
        "__or__": "or",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls.logical_operators.items():
            setattr(cls, meth, cls._get_logical_operator(opt))

    @classmethod
    def _get_logical_operator(cls, opt: str) -> Callable[["ScalarArray", Any], "ScalarArray"]:
        def func(self: "ScalarArray", other: Any) -> "BooleanScalarArray":
            if isinstance(other, SearchScalarContainer):
                expr = {f"${opt}": [self.expr, other.expr]}
            else:
                expr = {f"${opt}": [self.expr, other]}
            return BooleanScalarArray(expr, pt.boolean(), self.upper_expr)

        return func


class ScalarArray(SearchContainer):
    """One-dimensional array for portex builtin type array."""

    prefix = "$."

    def query(self, func: Callable[[Any], Any]) -> "ScalarArray":
        """Query the data of an ArraySeries with a lambda function.

        Arguments:
            func: The query function.

        Returns:
            The ArraySeries with the query expression.

        """
        expr = {
            "$filter": [
                self.upper_expr,
                func(
                    self.schema.search_container.item_container.from_upper(self.expr, self.schema)
                ).expr,
            ]
        }
        return ScalarArray(expr, self.schema, self.upper_expr)

    def any(self) -> "BooleanScalar":
        """Whether any element is True.

        Returns:
            The BooleanSeries with the any expression.

        """
        expr = {"$any_match": [self.upper_expr, self.expr]}
        return BooleanScalar(expr)

    def all(self) -> "BooleanScalar":
        """Whether all elements are True.

        Returns:
            The BooleanSeries with the all expression.

        """
        expr = {"$all_match": [self.upper_expr, self.expr]}
        return BooleanScalar(expr)


@SearchContainerRegister(pt.boolean)
class BooleanScalarArray(ScalarArray, LogicalOperatorsMixin):
    """One-dimensional array for portex builtin type array with the boolean items."""

    item_container = BooleanScalar


@SearchContainerRegister(pt.string)
class StringScalarArray(ScalarArray, LogicalOperatorsMixin):
    """One-dimensional array for portex builtin type array with the string and enum items."""

    item_container = StringScalar


@SearchContainerRegister(pt.enum)
class EnumScalarArray(ScalarArray):
    """One-dimensional array for portex builtin type array with the string and enum items."""

    item_container = EnumScalar

    def __eq__(self, other: Any) -> BooleanScalarArray:  # type: ignore[override]
        if isinstance(other, SearchScalarContainer):
            expr = {"$eq": [self.expr, other.expr]}
        else:
            values = self.schema.to_builtin().values  # type: ignore[attr-defined]
            if other in values:
                other = values.index(other)
            expr = {"$eq": [self.expr, other]}
        return BooleanScalarArray(expr, pt.boolean(), self.upper_expr)

    def __ne__(self, other: Any) -> BooleanScalarArray:  # type: ignore[override]
        if isinstance(other, SearchScalarContainer):
            expr = {"$ne": [self.expr, other.expr]}
        else:
            values = self.schema.to_builtin().values  # type: ignore[attr-defined]
            if other in values:
                other = values.index(other)
            expr = {"$ne": [self.expr, other]}
        return BooleanScalarArray(expr, pt.boolean(), self.upper_expr)


@SearchContainerRegister(
    pt.float32,
    pt.float64,
    pt.int32,
    pt.int64,
)
class NumberScalarArray(ScalarArray, ComparisonArithmeticOperatorsMixin):
    """One-dimensional array for portex builtin type array with the numerical items."""

    item_container = NumberScalar

    @classmethod
    def _get_comparison_operator(cls, opt: str) -> Callable[..., ScalarArray]:
        def func(self: "NumberScalarArray", other: Any) -> ScalarArray:
            if isinstance(other, SearchScalarContainer):
                self.check_type_for_other(other, opt)
                expr = {f"${opt}": [self.expr, other.expr]}
            else:
                expr = {f"${opt}": [self.expr, other]}

            return ScalarArray(expr, pt.array(pt.boolean()), self.upper_expr)

        return func

    @classmethod
    def _get_numeric_operator(cls, opt: str) -> Callable[..., "NumberScalarArray"]:
        def func(self: "NumberScalarArray", other: Any) -> "NumberScalarArray":
            if isinstance(other, SearchScalarContainer):
                self.check_type_for_other(other, opt)
                expr = {f"${opt}": [self.expr, other.expr]}
            else:
                expr = {f"${opt}": [self.expr, other]}

            return NumberScalarArray(expr, self.schema, self.upper_expr)

        return func

    def _check_upper_expr(self, opt: str) -> None:
        if self.expr != self.prefix:
            raise TypeError(f"{opt} operation only support for numerical array.")

    def size(self) -> "NumberScalar":
        """Get the length of array series.

        Returns:
            The NumberScalar with the size expression.

        """
        self._check_upper_expr("size")
        return NumberScalar({"$size": [self.upper_expr]}, self.schema)

    def max(self) -> "NumberScalar":
        """Get the max value of array series.

        Returns:
            The NumberScalar with the max expression.

        """
        self._check_upper_expr("max")
        return NumberScalar({"$max": [self.upper_expr]}, self.schema)

    def min(self) -> "NumberScalar":
        """Get the min value of array series.

        Returns:
            The NumberScalar with the min expression.

        """
        self._check_upper_expr("min")
        return NumberScalar({"$min": [self.upper_expr]}, self.schema)

    def sum(self) -> "NumberScalar":
        """Get the sum of array series.

        Returns:
            The NumberScalar with the sum expression.

        """
        self._check_upper_expr("sum")
        return NumberScalar({"$sum": [self.upper_expr]}, self.schema)


@SearchContainerRegister(pt.record)
class DataFrame(SearchContainer):
    """The Two-dimensional array for the search."""

    prefix = "$"
    item_container = RowSeries
    schema: pt.PortexRecordBase

    def __getitem__(self, key: str) -> SearchContainer:
        field = self.schema[key]
        expr = f"{self.expr}.{key}"
        return field.search_container(expr, field, self.upper_expr)

    def query(self, func: Callable[[Any], Any]) -> "DataFrame":
        """Query the data of an ArraySeries with a lambda function.

        Arguments:
            func: The query function.

        Returns:
            The DataFrame with the query expression.

        """
        expr = {"$filter": [self.upper_expr, func(RowSeries(self.schema)).expr]}
        return DataFrame(expr, self.schema, self.upper_expr)


class ArrayDistributor(SearchContainer):
    """A distributor to instance DataFrame, ArrayScalar by different array items."""

    @classmethod
    def from_upper(
        cls, expr: Union[str, Dict[str, Any]], schema: pt.PortexType
    ) -> "SearchContainer":
        """Instantiate a Search object from the upper level.

        Arguments:
            expr: The expression of the search.
            schema: The schema of the series.

        Returns:
            The loaded object.

        """
        items: pt.PortexType = schema.to_builtin().items  # type: ignore[attr-defined]
        return items.search_container.from_upper(expr, items)


@SearchContainerRegister(pt.array)
class ArraySeries(SearchContainer):
    """The One-dimensional array for the search."""

    item_container = ArrayDistributor



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""The implementation of the search related Scalar."""

from typing import Any, Callable, Dict, Union

import graviti.portex as pt
from graviti.dataframe.sql.container import SearchContainer, SearchScalarContainer

NUMERICAL_PRIORITY = {pt.int32: 0, pt.int64: 1, pt.float32: 2, pt.float64: 3}


class LogicalOperatorsMixin:
    """A mixin for dynamically implementing logical operators."""

    expr: Union[str, Dict[str, Any]]
    schema: pt.PortexType
    logical_operators: Dict[str, str] = {
        "__eq__": "eq",
        "__ne__": "ne",
        "__and__": "and",
        "__or__": "or",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls.logical_operators.items():
            setattr(cls, meth, cls._get_logical_operator(opt))

    @classmethod
    def _get_logical_operator(
        cls, opt: str
    ) -> Callable[["LogicalOperatorsMixin", Any], "BooleanScalar"]:
        def func(self: "LogicalOperatorsMixin", other: Any) -> "BooleanScalar":
            if isinstance(other, SearchScalarContainer):
                expr = {f"${opt}": [self.expr, other.expr]}
            else:
                expr = {f"${opt}": [self.expr, other]}
            return BooleanScalar(expr)

        return func


class ComparisonArithmeticOperatorsMixin:
    """A mixin for dynamically implementing comparison and arithmetic operators."""

    expr: Union[str, Dict[str, Any]]
    schema: pt.PortexType
    comparison_operators: Dict[str, str] = {
        "__gt__": "gt",
        "__ge__": "gte",
        "__lt__": "lt",
        "__le__": "lte",
    }
    numerical_opeartors: Dict[str, str] = {
        "__div__": "div",
        "__mod__": "mod",
        "__pow__": "pow",
        "__sub__": "sub",
        "__mul__": "mult",
        "__add__": "add",
    }

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        for meth, opt in cls.comparison_operators.items():
            setattr(cls, meth, cls._get_comparison_operator(opt))
        for meth, opt in cls.numerical_opeartors.items():
            setattr(cls, meth, cls._get_numeric_operator(opt))

    @classmethod
    def _get_comparison_operator(cls, opt: str) -> Callable[..., Any]:
        raise NotImplementedError

    @classmethod
    def _get_numeric_operator(cls, opt: str) -> Callable[..., Any]:
        raise NotImplementedError

    def check_type_for_other(self, other: SearchScalarContainer, opt: str) -> None:
        """Check whether the other series is the same type as self series.

        Arguments:
            other: The needed check series.
            opt: Name of opt.

        Raises:
            TypeError: When the right series type is different from left.

        """
        if not isinstance(other, self.__class__):
            raise TypeError(f"Invalid {opt} operation between {self.schema} and {other.schema}")


class NumberScalar(
    SearchScalarContainer,
    LogicalOperatorsMixin,
    ComparisonArithmeticOperatorsMixin,
):
    """One-dimensional array for numerical portex builtin type."""

    schema: Union[pt.int32, pt.int64, pt.float32, pt.float64]

    @classmethod
    def _get_comparison_operator(cls, opt: str) -> Callable[["NumberScalar", Any], "BooleanScalar"]:
        def func(self: "NumberScalar", other: Any) -> "BooleanScalar":
            if isinstance(other, SearchScalarContainer):
                self.check_type_for_other(other, opt)
                expr = {f"${opt}": [self.expr, other.expr]}
            else:
                expr = {f"${opt}": [self.expr, other]}

            return BooleanScalar(expr)

        return func

    @classmethod
    def _get_numeric_operator(cls, opt: str) -> Callable[["NumberScalar", Any], "NumberScalar"]:
        def func(self: "NumberScalar", other: Any) -> "NumberScalar":
            if isinstance(other, SearchScalarContainer):
                self.check_type_for_other(other, opt)
                expr = {f"${opt}": [self.expr, other.expr]}
            else:
                expr = {f"${opt}": [self.expr, other]}

            schema = (
                self.schema
                if NUMERICAL_PRIORITY[self.schema.__class__]
                > NUMERICAL_PRIORITY[other.schema.__class__]
                else other.schema
            )
            return cls(expr, schema)

        return func


class BooleanScalar(SearchScalarContainer, LogicalOperatorsMixin):
    """One-dimensional array for portex builtin type boolean."""

    def __init__(self, expr: Union[str, Dict[str, Any]]) -> None:
        super().__init__(expr, pt.boolean())


class StringScalar(SearchScalarContainer, LogicalOperatorsMixin):
    """One-dimensional array for portex builtin type string."""


class EnumScalar(SearchScalarContainer):
    """One-dimensional array for portex builtin type enum."""

    def __eq__(self, other: Any) -> BooleanScalar:  # type: ignore[override]
        if isinstance(other, SearchScalarContainer):
            expr = {"$eq": [self.expr, other.expr]}
        else:
            values = self.schema.to_builtin().values  # type: ignore[attr-defined]
            if other in values:
                other = values.index(other)
            expr = {"$eq": [self.expr, other]}
        return BooleanScalar(expr)

    def __ne__(self, other: Any) -> BooleanScalar:  # type: ignore[override]
        if isinstance(other, SearchScalarContainer):
            expr = {"$ne": [self.expr, other.expr]}
        else:
            values = self.schema.to_builtin().values  # type: ignore[attr-defined]
            if other in values:
                other = values.index(other)
            expr = {"$ne": [self.expr, other]}
        return BooleanScalar(expr)


class RowSeries(SearchScalarContainer):
    """The One-dimensional array for the search."""

    schema: pt.PortexRecordBase

    def __init__(self, schema: pt.PortexRecordBase) -> None:
        super().__init__("$", schema)

    def __getitem__(self, key: str) -> Union[SearchContainer, SearchScalarContainer]:
        field = self.schema[key]
        return field.search_container.item_container.from_upper(f"{self.expr}.{key}", field)



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Graviti file related class."""

from hashlib import sha1
from pathlib import Path
from typing import Dict, Type, TypeVar, Union
from urllib.request import url2pathname

import pyarrow as pa
from _io import BufferedReader

from graviti.utility.common import shorten
from graviti.utility.repr import ReprMixin
from graviti.utility.requests import UserResponse, config, get_session

_T = TypeVar("_T", bound="File")


class FileBase(ReprMixin):
    """This class represents the file in a DataFrame."""

    _url: str
    _checksum: str

    def open(self) -> Union[UserResponse, BufferedReader]:
        """Return the binary file pointer of this file.

        Raises:
            NotImplementedError: The method of the base class should not be called.

        """
        raise NotImplementedError

    @property
    def url(self) -> str:
        """Get the url of the file.

        Returns:
            The url of the file.

        """
        return self._url

    @property
    def checksum(self) -> str:
        """Get the checksum of the file.

        Returns:
            The checksum of the file.

        """
        return self._checksum

    @staticmethod
    def from_pyarrow(scalar: pa.StructScalar) -> Union["RemoteFile", "File"]:
        """Create File or RemoteFile instance with pyarrow struct scalar.

        Arguments:
            scalar: The input pyarrow struct scalar.

        Returns:
            The File or RemoteFile instance.

        """
        url: str = scalar["url"].as_py()
        checksum: str = scalar["checksum"].as_py()

        if url.startswith("file://"):
            return File._create(url, checksum)  # pylint: disable=protected-access

        return RemoteFile(url, checksum)

    def to_pyobj(self) -> Dict[str, str]:
        """Dump the local file to a python dict.

        Returns:
            A python dict representation of the local file.

        """
        return {"url": self.url, "checksum": self.checksum}


class RemoteFile(FileBase):
    """This class represents the file on Graviti platform.

    Arguments:
        url: The url string of the file.
        checksum: The checksum of the file.

    """

    def __init__(self, url: str, checksum: str) -> None:
        self._url = url
        self._checksum = checksum

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{shorten(self._checksum)}")'

    def open(self) -> Union[UserResponse, BufferedReader]:
        """Return the binary file pointer of this file.

        Returns:
            The remote file pointer.

        """
        session = get_session()
        return UserResponse(session.request("GET", self._url, timeout=config.timeout, stream=True))


class File(FileBase):
    """This class represents the local file.

    Arguments:
        path: The local path of the file.

    """

    _BUFFER_SIZE = 65536

    def __init__(self, path: str) -> None:
        self.path = Path(path).absolute()

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.path.name}")'

    @classmethod
    def _create(cls: Type[_T], url: str, checksum: str) -> _T:
        obj: _T = object.__new__(cls)
        obj.path = Path(url2pathname(url[7:]))
        obj._url = url
        obj._checksum = checksum
        return obj

    @property
    def url(self) -> str:
        """Get the url of the file.

        Returns:
            The url of the file.

        """
        if not hasattr(self, "_url"):
            self._url = self.path.as_uri()

        return self._url

    @property
    def checksum(self) -> str:
        """Get the sha1 checksum of the local file.

        Returns:
            The sha1 checksum of the local file.

        """
        if not hasattr(self, "_checksum"):
            sha1_object = sha1()
            with self.path.open("rb") as fp:
                while True:
                    data = fp.read(self._BUFFER_SIZE)
                    if not data:
                        break
                    sha1_object.update(data)

            self._checksum = sha1_object.hexdigest()

        return self._checksum

    def open(self) -> BufferedReader:
        """Return the binary file pointer of this file.

        Returns:
            The local file pointer.

        """
        return self.path.open("rb")



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#
"""The Portex builtin types."""


from functools import reduce
from operator import mul
from typing import Iterable, List, Mapping, Optional, Tuple, Type, TypeVar, Union

import pyarrow as pa

from graviti.portex import ptype as PTYPE
from graviti.portex.base import PortexRecordBase, PortexType
from graviti.portex.factory import ConnectedFieldsFactory
from graviti.portex.field import Fields
from graviti.portex.package import packages
from graviti.portex.param import Param, Params, param
from graviti.portex.register import PyArrowConversionRegister

builtins = packages.builtins
_E = Union[int, float, str, bool, None]


class PortexBuiltinType(PortexType):  # pylint: disable=abstract-method
    """The base class of Portex builtin type."""

    _T = TypeVar("_T", bound="PortexBuiltinType")

    params = Params()
    packages = builtins

    def __init_subclass__(cls) -> None:
        params = Params(cls.params)
        for name in getattr(cls, "__annotations__", {}):
            parameter = getattr(cls, name, None)
            if isinstance(parameter, tuple):
                params.add(Param(name, *parameter))
                delattr(cls, name)

        cls.params = params

        builtins[cls.__name__] = cls

    def __init__(self, nullable: bool = False) -> None:
        self.nullable = PTYPE.Boolean.check(nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], pyarrow_type: pa.DataType) -> _T:
        return cls()

    def to_builtin(self: _T) -> _T:
        """Expand the top level type to Portex builtin type.

        Returns:
            The expanded Portex builtin type.

        """
        return self


@PyArrowConversionRegister(pa.lib.Type_STRING)  # pylint: disable=c-extension-no-member
class string(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``string``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = string()
        >>> t
        string()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.string()


@PyArrowConversionRegister(pa.lib.Type_BINARY)  # pylint: disable=c-extension-no-member
class binary(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``binary``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = binary()
        >>> t
        binary()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.binary()


@PyArrowConversionRegister(pa.lib.Type_BOOL)  # pylint: disable=c-extension-no-member
class boolean(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``boolean``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = boolean()
        >>> t
        boolean()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.bool_()


@PyArrowConversionRegister(pa.lib.Type_INT32)  # pylint: disable=c-extension-no-member
class int32(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``int32``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = int32()
        >>> t
        int32()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.int32()


@PyArrowConversionRegister(pa.lib.Type_INT64)  # pylint: disable=c-extension-no-member
class int64(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``int64``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = int64()
        >>> t
        int64()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.int64()


@PyArrowConversionRegister(pa.lib.Type_FLOAT)  # pylint: disable=c-extension-no-member
class float32(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``float32``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = float32()
        >>> t
        float32()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.float32()


@PyArrowConversionRegister(pa.lib.Type_DOUBLE)  # pylint: disable=c-extension-no-member
class float64(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex primitive type ``float64``.

    Arguments:
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = float64()
        >>> t
        float64()

    """

    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        return pa.float64()


@PyArrowConversionRegister(pa.lib.Type_STRUCT)  # pylint: disable=c-extension-no-member
class record(PortexBuiltinType, PortexRecordBase):  # pylint: disable=invalid-name
    """Portex complex type ``record``.

    Arguments:
        fields: The fields of the record.
        nullable: Whether it is a nullable type.

    Examples:
        Create a record by dict:

        >>> t = record({"f0": int32(), "f1": float32()})
        >>> t
        record(
          fields={
            'f0': int32(),
            'f1': float32(),
          },
        )

        Create a record by tuple list:

        >>> t = record([("f0", string()), ("f1", enum(["v0", "v1"]))])
        >>> t
        record(
          fields={
            'f0': string(),
            'f1': enum(
              values=['v0', 'v1'],
            ),
          },
        )

    """

    _T = TypeVar("_T", bound="record")

    _fields_factory = ConnectedFieldsFactory.from_parameter_name("fields")

    fields: Fields = param(ptype=PTYPE.Fields)
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(
        self,
        fields: Union[Iterable[Tuple[str, PortexType]], Mapping[str, PortexType]],
        nullable: bool = False,
    ) -> None:
        self.fields = PTYPE.Fields.check(fields)
        super().__init__(nullable=nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], pyarrow_type: pa.StructType) -> _T:
        return cls((field.name, cls.from_pyarrow(field.type)) for field in pyarrow_type)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow struct DataType.

        """
        return pa.struct(self.fields.to_pyarrow())  # pylint: disable=no-member


class enum(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex complex type ``enum``.

    Arguments:
        values: The values of the enum members.
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = enum(["v0", "v1"])
        >>> t
        enum(
          values=['v0', 'v1'],
        )

    """

    values: List[_E] = param(ptype=PTYPE.Enum)
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(self, values: Iterable[_E], nullable: bool = False) -> None:
        self.values = PTYPE.Enum.check(values)
        super().__init__(nullable=nullable)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        length = len(self.values)
        if length < 128:
            return pa.int8()

        if length < 32768:
            return pa.int16()

        return pa.int32()


@PyArrowConversionRegister(
    pa.lib.Type_LIST, pa.lib.Type_FIXED_SIZE_LIST  # pylint: disable=c-extension-no-member
)
class array(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex complex type ``array``.

    Arguments:
        items: The item type of the array.
        length: The length of the array.
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = array(int32(0), 100)
        >>> t
        array(
          items=int32(
            minimum=0,
          ),
          length=100,
        )

    """

    _T = TypeVar("_T", bound="array")

    items: PortexType = param(ptype=PTYPE.PortexType)
    length: Optional[int] = param(None, ptype=PTYPE.Integer)
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(
        self, items: PortexType, length: Optional[int] = None, nullable: bool = False
    ) -> None:
        self.items = PTYPE.PortexType.check(items)
        self.length = PTYPE.Integer.check(length) if length is not None else None
        super().__init__(nullable=nullable)

    @classmethod
    def _from_pyarrow(cls: Type[_T], pyarrow_type: Union[pa.ListType, pa.FixedSizeListType]) -> _T:
        return cls(
            cls.from_pyarrow(pyarrow_type.value_type),
            getattr(pyarrow_type, "list_size", None),
        )

    def _get_column_count(self) -> int:
        """Get the total column count of the portex type.

        Returns:
            The total column count.

        """
        return self.items._get_column_count()  # pylint: disable=protected-access

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        list_size = self.length if self.length else -1
        return pa.list_(self.items.to_pyarrow(), list_size)  # pylint: disable=no-member


class tensor(PortexBuiltinType):  # pylint: disable=invalid-name
    """Portex complex type ``tensor``.

    Arguments:
        shape: The shape of the tensor.
        items: The item type of the tensor.
        nullable: Whether it is a nullable type.

    Examples:
        >>> t = tensor((3, 3), float64())
        >>> t
        tensor(
          shape=(3, 3),
          items=float64(),
        )

    """

    shape: Tuple[Optional[int], ...] = param(ptype=PTYPE.Array)
    items: PortexType = param(ptype=PTYPE.PortexType)
    nullable: bool = param(False, ptype=PTYPE.Boolean)

    def __init__(
        self, shape: Iterable[Optional[int]], items: PortexType, nullable: bool = False
    ) -> None:
        self.shape = tuple(PTYPE.Array.check(shape))
        self.items = PTYPE.PortexType.check(items)
        super().__init__(nullable=nullable)

    def to_pyarrow(self) -> pa.DataType:
        """Convert the Portex type to the corresponding builtin PyArrow DataType.

        Returns:
            The corresponding builtin PyArrow DataType.

        """
        try:
            list_size = reduce(mul, self.shape)
        except TypeError:
            list_size = -1
        return pa.list_(self.items.to_pyarrow(), list_size=list_size)  # pylint: disable=no-member



#!/usr/bin/env python3
#
# Copyright 2022 Graviti. Licensed under MIT License.
#

"""Common tools."""

from collections import defaultdict
from datetime import datetime
from functools import wraps
from threading import Lock
from typing import Any, Callable, DefaultDict, TypeVar

_CallableWithoutReturnValue = TypeVar("_CallableWithoutReturnValue", bound=Callable[..., None])

locks: DefaultDict[int, Lock] = defaultdict(Lock)


def urlnorm(url: str) -> str:
    """Normalized the input url by removing the trailing slash.

    Arguments:
        url: the url needs to be normalized.

    Returns:
        The normalized url.

    """
    return url.rstrip("/")


def locked(func: _CallableWithoutReturnValue) -> _CallableWithoutReturnValue:
    """The decorator to add threading lock for methods.

    Arguments:
        func: The method needs to add threading lock.

    Returns:
        The method with theading locked.

    """

    @wraps(func)
    def wrapper(self: Any, *arg: Any, **kwargs: Any) -> None:
        key = id(self)
        lock = locks[key]
        acquire = lock.acquire(blocking=False)
        try:
            if acquire:
                func(self, *arg, **kwargs)
                del locks[key]
            else:
                lock.acquire()
        finally:
            lock.release()

    return wrapper  # type: ignore[return-value]


def shorten(origin: str) -> str:
    """Return the first 7 characters of the original string.

    Arguments:
        origin: The string needed to be shortened.

    Returns:
        A string of length 7.

    """
    return origin[:7]


def convert_iso_to_datetime(iso: str) -> datetime:
    """Convert iso 8601 format string to datetime format time.

    Arguments:
        iso: The iso 8601 format string.

    Returns:
        The datetime format time.

    """
    return datetime.fromisoformat(iso.replace("Z", "+00:00"))
